from abc import ABC, abstractmethod
from enum import Enum, auto
from typing import Callable, List, Optional, Union

import pandas as pd
import torch

from .augmentation import Converter
from aidd_codebase.datamodules.tokenizer import Tokenizer


class DataType(Enum):
    SEQUENCE = auto()
    VALUE = auto()
    CLASS = auto()


class ReturnOptions(Enum):
    Molecule = auto()
    CLEANED = auto()
    Original = auto()


class _AbsProcessor(ABC):  # Should be protocol -> change
    def __init__(self, type: DataType) -> None:
        self.type = type

    @abstractmethod
    def load_data(self, data_path: str):
        pass

    @abstractmethod
    def set_data(self, data: Union[pd.DataFrame, pd.Series]) -> None:
        pass

    @abstractmethod
    def inspect_data(self):
        pass

    @abstractmethod
    def clean_data(self):
        pass

    @abstractmethod
    def augment_data(self, data: pd.DataFrame) -> pd.DataFrame:
        pass

    @abstractmethod
    def transform_data(self, data: pd.DataFrame) -> pd.DataFrame:
        pass

    @abstractmethod
    def clean_report(self):
        pass

    @abstractmethod
    def return_data(self):
        pass


class FloatProcessor(_AbsProcessor):
    def __init__(
        self,
        type: DataType,
        return_type: ReturnOptions,
        remove_missing: bool = True,
        constrains: Optional[List[Callable]] = None,
        augmentations: Optional[List[Callable]] = None,
    ) -> None:
        super().__init__(type)

        self.return_type = return_type
        assert isinstance(
            return_type, ReturnOptions
        ), "return_type should be in ReturnOptions class"

        self.remove_missing = remove_missing
        self.constrains = constrains
        self.augmentations = augmentations

    def load_data(self, data_path: str) -> None:
        self.data = pd.read_csv(data_path, header=None)

    def set_data(self, data: Union[pd.DataFrame, pd.Series]):
        self.data = data

    def inspect_data(self) -> None:
        print(self.data.head())

    def clean_data(self) -> None:
        self.original_len = len(self.data)
        self.original_len = len(self.data)

        cleaned_data = self.data.copy()

        # Missing
        if self.remove_missing:
            cleaned_data = cleaned_data.dropna()
            self.nmissing = self.original_len - len(cleaned_data)

        # Constrains
        if self.constrains:
            for constrain in self.constrains:
                cleaned_data = cleaned_data[constrain(cleaned_data)].dropna()
        self.nconstrain = (
            self.original_len - self.nmissing - len(cleaned_data)
            if self.constrains
            else 0
        )

        self.cleaned_data = cleaned_data

    def transform_data(self, data: pd.DataFrame) -> pd.DataFrame:
        return data.applymap(lambda x: torch.tensor(x))

    def augment_data(self, data: pd.DataFrame) -> pd.DataFrame:
        if self.augmentations:
            for augmentation in self.augmentations:
                data = augmentation(data)
        return data

    def clean_report(self) -> None:
        print("Clean Report:")
        print(f"\tOriginal: {self.original_len} datapoints, ", end="")
        print(f"Current: {len(self.cleaned_data)}")
        print(f"\tMissing: {self.nmissing}")
        print(f"\tConstrained: {self.nconstrain}")

    def return_data(self) -> pd.DataFrame:
        if self.return_type == ReturnOptions.CLEANED:
            return self.cleaned_data
        elif self.return_type == ReturnOptions.Original:
            return self.data.loc[self.cleaned_data.index, :]


class SmilesProcessor(_AbsProcessor):
    def __init__(
        self,
        type: DataType,
        return_type: ReturnOptions,
        tokenizer: Tokenizer,
        remove_duplicates: bool = True,
        remove_missing: bool = True,
        constrains: Optional[List[Callable]] = None,
        augmentations: Optional[List[Callable]] = None,
    ) -> None:
        super().__init__(type)

        self.return_type = return_type
        assert isinstance(
            return_type, ReturnOptions
        ), "return_type should be in ReturnOptions class"

        self.tokenizer = tokenizer
        self.remove_duplicates = remove_duplicates
        self.remove_missing = remove_missing
        self.constrains = constrains
        self.augmentations = augmentations

    def load_data(self, data_path: str) -> None:
        self.data = pd.read_csv(data_path, header=None)

    def set_data(self, data: Union[pd.DataFrame, pd.Series]):
        self.data = data

    def inspect_data(self) -> None:
        print(self.data.head())
        print("Length breakdown:", self.data.nunique(dropna=False))

    def clean_data(self) -> pd.DataFrame:
        self.original_len = len(self.data)
        # canonical_smiles = self.data.applymap(Converter.smile2canonical)
        canonical_smiles = self.data.copy()

        # Duplicates
        if self.remove_duplicates:
            canonical_smiles = canonical_smiles.drop_duplicates()
            self.nduplicates = self.original_len - len(canonical_smiles)
        else:
            self.nduplicates = 0

        # Missing
        if self.remove_missing:
            canonical_smiles = canonical_smiles.dropna()
            self.nmissing = (
                self.original_len - self.nduplicates - len(canonical_smiles)
            )
        else:
            self.nmissing = 0

        # Constrains
        if self.constrains is not None:
            for constrain in self.constrains:
                canonical_smiles = canonical_smiles[
                    constrain(canonical_smiles)
                ].dropna()
            self.nconstrain = (
                self.original_len
                - self.nduplicates
                - self.nmissing
                - len(canonical_smiles)
            )
        else:
            self.nconstrain = 0

        self.canonical_smiles = canonical_smiles

    def transform_data(self, data: pd.DataFrame) -> pd.DataFrame:
        return data.applymap(lambda x: self.tokenizer.smile_prep(x))

    def augment_data(self, data: pd.DataFrame) -> pd.DataFrame:
        if self.augmentations:
            original = len(data)
            print(len(data))
            for augmentation in self.augmentations:
                data = augmentation(data)
            print(
                f"\tAugmentation: from {original} datapoints,"
                + f"to {len(data)} datapoints."
            )
        return data

    def clean_report(self) -> None:
        print("Clean Report:")
        print(f"\tOriginal: {self.original_len} datapoints, ", end="")
        print(f"Current: {len(self.canonical_smiles)}")
        print(f"\tDuplicates: {self.nduplicates}")
        print(f"\tMissing: {self.nmissing}")
        print(f"\tConstrained: {self.nconstrain}")

    def return_data(self) -> pd.DataFrame:
        if self.return_type == ReturnOptions.CLEANED:
            return self.canonical_smiles
        elif self.return_type == ReturnOptions.Original:
            return self.data.loc[self.canonical_smiles.index, :]
        elif self.return_type == ReturnOptions.Molecule:
            return self.data.loc[self.canonical_smiles.index, :].applymap(
                Converter.smile2mol
            )


# ==============
# 1.1 Inspection
# ==============

# Detect unexpeted, incorrect or inconsistent data.

# Data profiling (summary statistics)
# Visualizations (outlier detection, ranges stds and the unexpected)


# ============
# 1.2 Cleaning
# ============

# Fix or remove the anomalies discovered.

# Irrelevant data: remove data unnessecary to our current goals
# Duplicates: combine or remove based on identical identifiers
# Type conversion: make sure tensors are tensors, string are strings etc.
# Syntax errors: padding remove/add, fix typos (nspectfor low-frequence values)
# Standardize: all units the same, all strings should have similar syntax
# Scaling/Transformation: Skaling 0-100 to 0-1
# Normalization: nomrally distributed data
# Missing Data: Remove all missing i.e. NA, 0, Null, etc., flag or impute
# Outliers: all values outside the 1.5 sigma away should be considered outliers
#           (but not neccessarily guilty)


# =============
# 1.3 Verifying
# =============

# After cleaning, ther results should be verified for correctness

# Redo Inspection en checks
# Verify that the data can be split correctly


# =============
# 1.4 Reporting
# =============

# A report about made changes and the qualitify of current data should be made

# Report all alterations, removals and final statistics of the dataset



import numpy as np
from torch import optim


# class SchedulerProtocol(Protocol):
#     def get_lr(self) -> float:
#         """Returns the next learning rate."""

#     def step(self, epoch: int) -> None:
#         """Instantiates the optimizer with the lr of the epoch"""


class Scheduler:  # optim.lr_sheduler._LRScheduler
    pass


class LinearPiecewiseScheduler(Scheduler):
    pass


class CosineWarmupScheduler(Scheduler):
    def __init__(self, optimizer, warmup, max_iters):
        self.warmup = warmup
        self.max_num_iters = max_iters
        super().__init__(optimizer)

    def step(self, epoch: int) -> None:
        return self.get_lr(epoch=epoch)

    def get_lr(self, epoch: int):
        lr_factor = self.get_lr_factor(epoch=epoch)
        return [base_lr * lr_factor for base_lr in self.base_lrs]

    def get_lr_factor(self, epoch):
        lr_factor = 0.5 * (1 + np.cos(np.pi * epoch / self.max_num_iters))
        if epoch <= self.warmup:
            lr_factor *= epoch * 1.0 / self.warmup
        return lr_factor



from typing import Optional

import torch.nn as nn

from aidd_codebase.utils.typescripts import Tensor


class WeightSharingEncoder(nn.TransformerEncoder):
    def __init__(
        self,
        encoder_layer: nn.Module,
        num_layers: int,
        norm: Optional[nn.Module],
    ) -> None:
        super().__init__(encoder_layer, num_layers=1, norm=norm)
        self.total_layers = num_layers  # prevent conflict with super()

    def forward(
        self,
        src: Tensor,
        mask: Optional[Tensor] = None,
        src_key_paddin_mask: Optional[Tensor] = None,
    ):
        x = src
        for _ in range(self.total_layers):
            x = self.layers[0](
                x,
                mask,
                src_key_paddin_mask,
            )

        if self.norm:
            x = self.norm(x)

        return x


class WeightSharingDecoder(nn.TransformerDecoder):
    def __init__(
        self,
        decoder_layer: nn.Module,
        num_layers: int,
        norm: Optional[nn.Module],
    ) -> None:
        super().__init__(decoder_layer, num_layers=1, norm=norm)
        self.total_layers = num_layers  # prevent conflict with super()

    def forward(
        self,
        tgt: Tensor,
        memory: Tensor,
        tgt_mask: Optional[Tensor] = None,
        memory_mask: Optional[Tensor] = None,
        tgt_key_padding_mask: Optional[Tensor] = None,
        memory_key_padding_mask: Optional[Tensor] = None,
    ):
        x = tgt
        for _ in range(self.total_layers):
            x = self.layers[0](
                x,
                memory,
                tgt_mask,
                memory_mask,
                tgt_key_padding_mask,
                memory_key_padding_mask,
            )

        if self.norm:
            x = self.norm(x)

        return x



import functools
from typing import Any, Callable, Dict


def compose(*functions: Callable) -> Callable:
    """Helper function to compose together sequential operations."""
    return functools.reduce(lambda f, g: lambda x: g(f(x)), functions)


def convert_dict_names_to_lower(dict: Dict[str, Any]) -> Dict[str, Any]:
    return {k.lower(): v for k, v in dict.items()}


def convert_dict_names_to_upper(dict: Dict[str, Any]) -> Dict[str, Any]:
    return {k.upper(): v for k, v in dict.items()}



from typing import List, Optional, Union

import pandas as pd
from rdkit import Chem


class Enumerator:
    def __init__(
        self,
        enumerations: int,
        seed: int,
        oversample: int = 0,
        max_len: Optional[int] = None,
        keep_original: bool = True,
    ) -> None:
        self.enumerations = enumerations
        self.oversample = oversample
        self.max_len = max_len
        self.keep_original = keep_original

        self.seed = seed

    def smiles_enumeration(
        self,
        smile: str,
    ) -> Union[List[str], None]:
        try:
            enum_smiles = Chem.MolToRandomSmilesVect(
                Chem.MolFromSmiles(smile),
                self.enumerations + self.oversample,
                randomSeed=self.seed,
            )
        except Exception:
            return None

        if self.keep_original:
            unique_enum_smiles = list(set([smile, *enum_smiles]))
            max_n = self.enumerations + 1
        else:
            unique_enum_smiles = list(set(enum_smiles))
            max_n = self.enumerations

        if self.max_len:
            unique_enum_smiles = [
                smile
                for smile in unique_enum_smiles
                if len(smile) <= self.max_len
            ]

        return (
            unique_enum_smiles[:max_n]
            if len(unique_enum_smiles) >= max_n
            else None
        )

    def dataframe_enumeration(self, data: pd.DataFrame) -> pd.DataFrame:
        data = data.applymap(lambda x: self.smiles_enumeration(x))
        data = data.dropna()
        return data.apply(pd.Series.explode)


class Converter:
    @staticmethod
    def smile2mol(smile: str) -> Chem:
        return Chem.MolFromSmiles(smile)

    @staticmethod
    def mol2canonical(mol: Chem) -> Union[str, None]:
        try:
            return Chem.MolToSmiles(
                mol,
                isomericSmiles=True,
                kekuleSmiles=False,
                canonical=True,
                allBondsExplicit=False,
                allHsExplicit=False,
            )
        except Exception:
            return None

    @staticmethod
    def smile2canonical(smile: str) -> Union[str, None]:
        try:
            return Chem.MolToSmiles(
                Chem.MolFromSmiles(smile),
                isomericSmiles=True,
                kekuleSmiles=False,
                canonical=True,
                allBondsExplicit=False,
                allHsExplicit=False,
            )
        except Exception:
            return None



import pickle
from pathlib import Path
from typing import Any

import pytorch_lightning as pl
import torch
import torch.nn as nn


def save_model(
    model: nn.Module, name: str, path: str, save_weights: bool = True
) -> None:
    Path(path).mkdir(parents=True, exist_ok=True)
    torch.save(model, f"{path}/{name}.pth")
    if save_weights:
        torch.save(model.state_dict(), f"{path}/{name}_weights.pth")


def load_model_weights(model: nn.Module, weight_path: str) -> nn.Module:
    return model.load_state_dict(torch.load(weight_path))


def load_model(path: str) -> nn.Module:
    return torch.load(path)


def checkpoint_load(
    object: pl.LightningModule, checkpoint_path: str
) -> pl.LightningModule:
    return object.load_from_checkpoint(checkpoint_path)


def pickle_save(file_path: str, object: Any):
    with open(file_path, "wb") as handle:
        pickle.dump(object, handle, protocol=pickle.HIGHEST_PROTOCOL)


def pickle_load(file_path: str) -> Any:
    with open(file_path, "rb") as handle:
        object = pickle.load(handle)
    return object



import warnings
from abc import abstractclassmethod
from enum import Enum, auto
from typing import Any, Callable, Dict, Optional, Tuple

from .config import _ABCDataClass


class Singleton:
    _instance = None

    def __new__(cls, *args, **kwargs):
        if cls._instance is None:
            cls._instance = super().__new__(cls, *args, **kwargs)
        return cls._instance


class Borg:
    _dict = None

    def __new__(cls, *args, **kwargs):
        obj = super().__new__(cls, *args, **kwargs)
        if cls._dict is None:
            cls._dict = obj.__dict__
        else:
            obj.__dict__ = cls._dict
        return obj


class CreditType(Enum):
    REFERENCE = auto()
    ACKNOWLEDGEMENT = auto()
    NONE = auto()


class Accreditation(Singleton):
    accreditations: Dict[
        str,
        Tuple[Tuple[Optional[str], Optional[str], Optional[str]], CreditType],
    ] = {}
    called_creditations: Dict[
        str,
        Tuple[Tuple[Optional[str], Optional[str], Optional[str]], CreditType],
    ] = {}

    def register_accreditation(
        self,
        call_name: str,
        author: str = None,
        github_handle: str = None,
        additional_information: str = None,
        credit_type: CreditType = CreditType.NONE,
    ) -> None:
        self.accreditations[call_name] = (
            (author, github_handle, additional_information),
            credit_type,
        )

    def get_accreditation(
        self, call_name: str
    ) -> Tuple[Tuple[Optional[str], Optional[str], Optional[str]], CreditType]:
        return self.accreditations[call_name]

    def notice_accreditation(self, call_name: str) -> None:
        if call_name in self.called_creditations.keys():
            pass
        else:
            try:
                self.called_creditations[call_name] = self.get_accreditation(
                    call_name
                )
            except Exception:
                pass

    def choice_called(self, call_name: str) -> None:
        self.notice_accreditation(call_name)

    def return_accreditations(
        self,
    ) -> Dict[
        str,
        Tuple[Tuple[Optional[str], Optional[str], Optional[str]], CreditType],
    ]:
        return self.called_creditations


class DictChoiceFactory:
    """Factory to generate a dictionary choice class"""

    dict_choice: Dict[str, Callable] = {}
    choice_arguments: Dict[str, _ABCDataClass] = {}
    accreditation: Accreditation = Accreditation()

    @classmethod
    def register_choice(
        cls,
        call_name: str,
        author: Optional[str] = None,
        github_handle: Optional[str] = None,
        additional_information: Optional[str] = None,
        credit_type: CreditType = CreditType.NONE,
    ) -> Callable:
        """Registers choices by string with a decorator.
        Optional: registers accreditation.
        """

        def wrapper(callable_cls: Any) -> Any:
            if author or github_handle or additional_information:
                cls.accreditation.register_accreditation(
                    call_name,
                    author,
                    github_handle,
                    additional_information,
                    credit_type,
                )
            cls.dict_choice[call_name] = callable_cls
            return callable_cls

        return wrapper

    @classmethod
    def register_prebuilt_choice(
        cls,
        call_name: str,
        callable_cls: Any,
        author: Optional[str] = None,
        github_handle: Optional[str] = None,
        additional_information: Optional[str] = None,
        credit_type: CreditType = CreditType.NONE,
    ) -> Any:
        """Registers choices by string directly.
        Optional: registers accreditation.
        """
        if author or github_handle or additional_information:
            cls.accreditation.register_accreditation(
                call_name,
                author,
                github_handle,
                additional_information,
                credit_type,
            )
        cls.dict_choice[call_name] = callable_cls

    @classmethod
    def register_arguments(cls, call_name: str) -> Callable:
        def wrapper(argument_class: _ABCDataClass) -> Any:
            cls.choice_arguments[call_name] = argument_class
            return argument_class

        return wrapper

    @classmethod
    def get_arguments(cls, call_name: str) -> _ABCDataClass:
        return cls.choice_arguments[call_name]

    @abstractclassmethod
    def get_choice(cls, call_name: str) -> Callable:
        """Returns a choice pointer out of the dict_choice dictionary."""
        cls.accreditation.choice_called(call_name)
        return cls.dict_choice[call_name]

    @classmethod
    def check_choice(cls, call_name: str) -> bool:
        """Checks if a choice is valid and returns a bool."""
        if call_name not in cls.dict_choice.keys():
            warnings.warn(f"{call_name} is not recognized.")
            return False
        return True

    @classmethod
    def validate_choice(cls, call_name: str) -> None:
        """Checks if a choice is valid and stops if not."""
        if call_name not in cls.dict_choice.keys():
            raise ValueError(f"{call_name} is not recognized.")

    @classmethod
    def get_accreditations(
        cls,
    ) -> Dict[
        str,
        Tuple[Tuple[Optional[str], Optional[str], Optional[str]], CreditType],
    ]:
        return cls.accreditation.return_accreditations()

    @classmethod
    def view_accreditations(
        cls,
    ) -> None:
        creditations = cls.accreditation.return_accreditations()
        for key, value in creditations.items():
            print(
                f"Object {key}: Author - {value[0][0]}, Github - {value[0][1]}"
                + f", Info - {value[0][2]}\n\t Credit Type {value[1]}"
            )



from dataclasses import dataclass, field
from typing import List

import torch
import torchmetrics


@dataclass
class Metric:
    name: str = "metric"
    values: List[float] = field(default_factory=list)
    running_total: float = 0.0
    num_updates: float = 0.0
    average: float = 0.0

    def update(self, value: float, size: int = 1):
        self.values.append(value)
        self.running_total += value * size
        self.num_updates += size
        # weighted average
        self.average = self.running_total / self.num_updates


# example of RMSE in a torch (lightning) metric #TODO
class TorchMetric(torchmetrics.Metric):
    def __init__(self, name: str) -> None:
        super().__init__()
        self.add_state("values", torch.tensor(0), dist_reduce_fx="sum")
        self.add_state("n_observations", torch.tensor(0), dist_reduce_fx="sum")

    def update(self, preds, target):
        self.values += torch.sum((preds - target) ** 2)
        self.n_observations += preds.numel()

    def cumpute(self):
        return torch.sqrt(self.RMSE / self.n_observations)


# Example of usage of torchmetrics
# def __init__(self):
#     ...
#     self.train_acc = torchmetrics.Accuracy()
#     self.valid_acc = torchmetrics.Accuracy()

# def training_step(self, batch, batch_idx):
#     x, y = batch
#     preds = self(x)
#     ...
#     self.train_acc(preds,y)
#     self.log('train_acc', self.train_acc, on_step=True, on_epoch=False)
#     ...

# def validation_step(self, batch, batch_idx):
#     x, y = batch
#     logits = self(x)
#     ...
#     self.valid_acc(logits, y)
#     self.log('valid_acc', self.valid_acc, on_step=True, on_epoch=True)



from dataclasses import dataclass
from typing import Optional

import pytorch_lightning as pl
import torch.nn as nn

from ..utils.config import _ABCDataClass
from ..utils.metacoding import CreditType
from ..utils.typescripts import Tensor
from .embedding.embedding import TokenEmbedding
from .embedding.positional import SequencePositionalEncoding
from .modelchoice import ModelChoice
from .modules.modules import WeightSharingDecoder


@ModelChoice.register_arguments(call_name="pl_decoder")
@dataclass(unsafe_hash=True)
class DecoderArguments(_ABCDataClass):
    tgt_vocab_size = 112
    num_decoder_layers = 3
    emb_size = 512
    num_heads = 8
    dim_feedforward = 512
    dropout: float = 0.1
    weight_sharing: bool = False
    max_seq_len: Optional[int] = None


@ModelChoice.register_choice(
    call_name="pl_decoder",
    author="Peter Hartog",
    github_handle="PeterHartog",
    credit_type=CreditType.NONE,
)
class Decoder(pl.LightningModule):
    def __init__(
        self,
        model_args: DecoderArguments,
    ) -> None:
        super().__init__()

        self.positional_encoding = SequencePositionalEncoding(
            emb_size=model_args.emb_size,
            dropout=model_args.dropout,
            maxlen=model_args.max_seq_len,
        )
        self.tgt_tok_emb = TokenEmbedding(
            vocab_size=model_args.tgt_vocab_size, emb_size=model_args.emb_size
        )
        norm = nn.LayerNorm(model_args.emb_size)
        decoder_layer = nn.TransformerDecoderLayer(
            d_model=model_args.emb_size,
            nhead=model_args.num_heads,
            dim_feedforward=model_args.dim_feedforward,
            dropout=model_args.dropout,
            activation="relu",
        )

        if model_args.weight_sharing:
            self.decoder = WeightSharingDecoder(
                decoder_layer=decoder_layer,
                num_layers=model_args.num_decoder_layers,
                norm=norm,
            )
        else:
            self.decoder = nn.TransformerDecoder(
                decoder_layer=decoder_layer,
                num_layers=model_args.num_decoder_layers,
                norm=norm,
            )

    # (ignore mypy error using type: ignore)
    def forward(  # type: ignore
        self,
        tgt: Tensor,
        memory: Tensor,
        tgt_mask: Optional[Tensor] = None,
        memory_mask: Optional[Tensor] = None,
        tgt_padding_mask: Optional[Tensor] = None,
        memory_padding_mask: Optional[Tensor] = None,
    ) -> Tensor:
        tgt_embedded = self.positional_encoding(self.tgt_tok_emb(tgt))
        return self.decoder(
            tgt=tgt_embedded,
            memory=memory,
            tgt_mask=tgt_mask,
            memory_mask=memory_mask,
            tgt_key_padding_mask=tgt_padding_mask,
            memory_key_padding_mask=memory_padding_mask,
        )



from typing import List

import pytorch_lightning as pl
import torch
import torch.nn as nn

from aidd_codebase.utils.typescripts import Tensor


class Max(nn.Module):
    def __init__(self, dim: int) -> None:
        super().__init__()
        self.dim = dim

    def forward(self, x: Tensor) -> Tensor:
        return torch.max(x, dim=self.dim).values


class KernelUnit(pl.LightningModule):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: int,
        stride: int,
    ) -> None:
        super().__init__()

        self.kernel = nn.Sequential(
            nn.Conv1d(
                in_channels=in_channels,
                out_channels=out_channels,
                kernel_size=kernel_size,
                stride=stride,
            ),
            nn.ReLU(),
            Max(dim=2),
        )

    def forward(self, x: Tensor) -> Tensor:
        self.kernel.to(x.device)
        return self.kernel(x)


class TextCNN(pl.LightningModule):
    def __init__(
        self,
        kernel_sizes: List[int],
        filters: List[int],
        stride: int,
        in_size: int,
        out_size: int,
        dropout: float = 0.1,
    ) -> None:
        super().__init__()
        self.in_size = in_size
        self.stride = stride
        self.out_size = out_size
        self.dropout = dropout

        # Create Convolution and Pooling Layers
        self.kernel_sizes = kernel_sizes
        self.filters = filters
        self.kernel_layers = [
            KernelUnit(
                in_channels=self.in_size,
                out_channels=filter,
                kernel_size=kernel_size,
                stride=self.stride,
            )
            for kernel_size, filter in zip(self.kernel_sizes, self.filters)
        ]

        self.output_net = nn.Sequential(
            nn.Linear(sum(self.filters), self.out_size),
            nn.Dropout(self.dropout),
            nn.Sigmoid(),
        )

    def forward(self, x: Tensor) -> Tensor:
        x = x.permute(1, 0, 2)  # Batch First

        concatenated_kernels = torch.cat(
            [kernel_layer(x) for kernel_layer in self.kernel_layers],
            dim=-1,
        )
        return self.output_net(concatenated_kernels)



import os
import warnings
from dataclasses import dataclass, field, fields, make_dataclass
from types import MappingProxyType
from typing import Any, Dict, List, Tuple

import yaml
from aidd_codebase import __version__


@dataclass(unsafe_hash=True)
class _ABCDataClass:
    def __reduce__(self):
        state = self.__dict__.copy()
        return (
            _NestedClassGetter(),
            (
                _ABCDataClass,
                self.__class__.__name__,
            ),
            state,
        )


class _NestedClassGetter(object):
    """
    When called with the containing class as the first argument,
    and the name of the nested class as the second argument,
    returns an instance of the nested class.
    """

    def __call__(self, containing_class, class_name):
        nested_class = getattr(containing_class, class_name)

        # make an instance of a simple object (this one will do), for which we
        # can change the __class__ later on.
        nested_instance = _NestedClassGetter()

        # set the class of the instance, the __init__ will never be called on
        # the class but the original state will be set later on by pickle.
        nested_instance.__class__ = nested_class
        return nested_instance


@dataclass(unsafe_hash=True)
class Config:
    codebase_version: str = __version__
    _dataclasses: Dict[str, _ABCDataClass] = field(default_factory=lambda: {})

    def dataclass_config_override(
        self, data_classes: Dict[str, _ABCDataClass]
    ) -> None:
        for k, data_class in data_classes.items():
            self._dataclasses[k] = data_class  # original Dataclasses
            self.__dict__[k] = data_class.__dict__  # mutable objects

    def _read_config_yaml(self, file: str) -> Dict:
        with open(file) as file:
            yaml_config = yaml.safe_load(file)
        return yaml_config

    def yaml_config_override(self, file: str) -> None:
        if os.path.exists(file) and self._read_config_yaml(file):
            yaml_file = self._read_config_yaml(file)
            for argument, group in yaml_file.items():
                if not isinstance(group, Dict):
                    warnings.warn(f"{group} not in dictionary.")
                else:
                    if argument in self._dataclasses.keys():
                        dataclass_fields = [
                            field.name
                            for field in fields(self._dataclasses[argument])
                        ]
                        for k in group.keys():
                            if k not in dataclass_fields:
                                warnings.warn(f"Unknown key found: {k}.")

                        group = {
                            k: v
                            for k, v in group.items()
                            if k in dataclass_fields
                        }
                        self._dataclasses[argument] = self._dataclasses[
                            argument
                        ](**group)
                        self.__dict__[argument] = self._dataclasses[
                            argument
                        ].__dict__
                        for k, v in group.items():
                            if (
                                k
                                not in self._dataclasses[
                                    argument
                                ].__dict__.keys()
                            ):
                                warnings.warn(f"Unknown key found: {k}.")
                    else:
                        self.__dict__[argument] = group

    def print_dataclass_arguments(self, data_class: _ABCDataClass) -> None:
        for k, v in data_class.items():
            if not k.startswith("__"):
                print(f"Arg {k} = {v}")

    def print_arguments(self) -> None:
        """Prints all arguments in a dataclass."""
        for k, v in self.__dict__.items():
            if not isinstance(v, (MappingProxyType, Dict)):
                if not k.startswith("__"):
                    print(f"Arg {k} = {v}")
            else:
                self.print_dataclass_arguments(v)

    def return_dataclass(self, call_name: str) -> dataclass:
        def choose_field(key: str, field_value: Any) -> Tuple:
            if isinstance(field_value, (Dict, List)):
                return (
                    key,
                    "typing.Any",
                    field(default_factory=lambda: field_value),
                )
            elif field_value is not None:
                return (key, "typing.Any", field(default=field_value))
            else:
                return (key, "typing.Optional", None)

        dataclass_fields = [
            choose_field(key, field_value)
            for key, field_value in self.__dict__[call_name].items()
            if not key.startswith("__")
        ]
        new_dataclass = make_dataclass(call_name, dataclass_fields)
        return new_dataclass()



import torch.nn as nn

from aidd_codebase.utils.metacoding import DictChoiceFactory
from aidd_codebase.utils.typescripts import Tensor


class ActivationChoice(DictChoiceFactory):
    pass


# Registers choices for loss functions of prebuilt loss functions
ActivationChoice.register_prebuilt_choice(
    call_name="relu", callable_cls=nn.ReLU
)
ActivationChoice.register_prebuilt_choice(
    call_name="kullback_leibler_div", callable_cls=nn.KLDivLoss
)
ActivationChoice.register_prebuilt_choice(
    call_name="kullback_leibler_recon_div", callable_cls=nn.KLRecDivLoss
)


@ActivationChoice.register_choice("custom_relu")
class ReLU(nn.Module):
    def __init__(self) -> None:
        super().__init__()

    def forward(x: Tensor) -> Tensor:
        return x.clamp_min(0.0).requires_grad_(True)


@ActivationChoice.register_choice("relu_alt")
class AlternativeReLU(nn.Module):
    def __init__(self) -> None:
        super().__init__()

    def forward(x: Tensor) -> Tensor:
        return (
            x.clamp_min(0.0).requires_grad_(True) - 0.5
        )  # To get better distr after relu



from dataclasses import dataclass
from datetime import datetime
from pathlib import Path


def validate_or_create_dir(dir_path: str) -> str:
    Path(dir_path).mkdir(parents=True, exist_ok=True)
    return dir_path


@dataclass
class Directories:
    """Creates some generic directories for a specific run."""

    PROJECT: str
    HOME_DIR: str
    TIME: str = datetime.now().strftime("%H:%M")
    DATE: str = datetime.now().strftime("%Y-%m-%d")

    def __post_init__(self):
        self.DATE_TIME = f"{self.DATE}/{self.TIME}"
        self.RUN_DIR: str = self.dir(f"{self.HOME_DIR}/runs/{self.PROJECT}")
        self.LOG_DIR: str = self.dir(f"{self.RUN_DIR}/logs")
        self.MODEL_DIR: str = self.dir(
            f"{self.RUN_DIR}/models/{self.DATE_TIME}"
        )
        self.CHECKPOINT_DIR: str = self.dir(
            f"{self.RUN_DIR}/checkpoint/{self.DATE_TIME}"
        )
        self.WEIGHTS_DIR: str = self.dir(
            f"{self.RUN_DIR}/weights/{self.DATE_TIME}"
        )

    @staticmethod
    def dir(dir: str) -> str:
        return validate_or_create_dir(dir)



import pytorch_lightning as pl
import torch.nn as nn

from aidd_codebase.utils.typescripts import Tensor


class _HighwayUnit(pl.LightningModule):
    """Abstract concept of a highway unit where T(x) is the transform gate
    and H(x) is the transform.
    """

    def __init__(
        self,
        in_features: int,
        out_features: int,
    ) -> None:
        super().__init__()
        self.H = nn.Sequential(
            nn.Linear(in_features, out_features),
            nn.ReLU(),
        )
        self.T = nn.Sequential(
            nn.Linear(in_features, out_features),
            nn.Sigmoid(),
        )

    def forward(self, x: Tensor) -> Tensor:
        transform_gate = self.T(x)  # prevent double backprop?
        y = transform_gate * self.H(x) + x * (1 - transform_gate)
        return y


class HighwayUnit(pl.LightningModule):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: int,
        stride: int,
    ) -> None:
        super().__init__()
        self.conv1 = nn.Conv1d(
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=kernel_size,
            stride=stride,
        )
        self.conv2 = nn.Conv1d(
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=kernel_size,
            stride=stride,
        )
        self.relu1 = nn.ReLU()
        self.relu2 = nn.ReLU()

    def forward(self, x: Tensor) -> Tensor:
        y = x + self.relu1(self.conv1(x)) * self.relu2(self.conv2(x))
        return y


class ConstantScalingUnit(pl.LightningModule):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: int,
        stride: int,
    ) -> None:
        super().__init__()
        self.conv1 = nn.Conv1d(
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=kernel_size,
            stride=stride,
        )
        self.conv2 = nn.Conv1d(
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=kernel_size,
            stride=stride,
        )
        self.relu1 = nn.ReLU()
        self.relu2 = nn.ReLU()

    def forward(self, x: Tensor) -> Tensor:
        y = 0.5 * x + 0.5 * self.relu1(self.conv1(x)) * self.relu2(
            self.conv2(x)
        )
        return y


class DropoutShortcutUnit(pl.LightningModule):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: int,
        stride: int,
        dropout: float = 0.1,
    ) -> None:
        super().__init__()
        self.conv1 = nn.Conv1d(
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=kernel_size,
            stride=stride,
        )
        self.conv2 = nn.Conv1d(
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=kernel_size,
            stride=stride,
        )
        self.relu1 = nn.ReLU()
        self.relu2 = nn.ReLU()
        self.dropout = nn.Dropout(p=dropout)

    def forward(self, x: Tensor) -> Tensor:
        y = self.dropout(x) + self.relu1(self.conv1(x)) * self.relu2(
            self.conv2(x)
        )
        return y


class ConvolutionShortcutUnit(pl.LightningModule):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: int,
        stride: int,
    ) -> None:
        super().__init__()
        self.conv1 = nn.Conv1d(
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=kernel_size,
            stride=stride,
        )
        self.conv2 = nn.Conv1d(
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=kernel_size,
            stride=stride,
        )
        self.conv3 = nn.Conv1d(
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=kernel_size,
            stride=stride,
        )
        self.relu1 = nn.ReLU()
        self.relu2 = nn.ReLU()
        self.relu3 = nn.ReLU()

    def forward(self, x: Tensor) -> Tensor:
        y = self.relu3(self.conv3(x)) + self.relu1(self.conv1(x)) * self.relu2(
            self.conv2(x)
        )
        return y



# from pytorch_lightning.loops import FitLoop
from typing import Any

import pytorch_lightning as pl
from aidd_codebase.framework.loopchoice import LoopChoice
from aidd_codebase.utils.metacoding import CreditType
from aidd_codebase.utils.typescripts import Tensor


@LoopChoice.register_choice(
    call_name="pl_seq2seq_loops",
    author="Peter Hartog",
    github_handle="PeterHartog",
    credit_type=CreditType.NONE,
)
class Seq2SeqLoop(pl.LightningModule):
    """Performs the training step, assumed dims [L, B]"""

    def __init__(self) -> None:
        super().__init__()

    def basic_loop(self, batch: Any, batch_idx: int, stage: str) -> Tensor:
        # x, y = batch                      moved to training_step
        # y_hat = model(x)                  moved to training_step
        # loss = loss_function(y_hat, y)    moved to training_step
        pass

    @staticmethod
    def translation(self, batch, batch_idx, stage):
        x, y = batch
        batch_size = x.shape[1]  # assumed dims are [L, B]

        tgt_in = y[:-1, :]
        logits = self(x, tgt_in)

        loss_key = self.get_loss_key(stage)
        for i, key in enumerate(self.metrics_list[stage].keys()):
            if key == loss_key:
                self.metrics_list[stage][key].update(
                    self.loss(logits, y[1:, :]), batch_size
                )
            else:
                self.metrics_list[stage][key].update(
                    self.metrics[i](logits, y[1:, :]), batch_size
                )

        outputs = {loss_key: self.metrics_list[stage][loss_key].values[-1]}
        return outputs

    @staticmethod
    def regression(self, batch, batch_idx):
        x, y = batch
        batch_size = x.shape[1]  # assumed dims are [L, B]

        y_hat = self(x)
        y_hat = y_hat.squeeze(1)

        for key in self.metrics_list.keys():
            self.train_metrics[key].update(
                self.metrics[key](y, y_hat), batch_size
            )

        outputs = {
            "loss": self.train_metrics["loss"].values[-1],
            "batch_size": batch_size,
        }
        return outputs

    @staticmethod
    def classification(self, batch, batch_idx):
        return self.regression(batch, batch_idx)


# class MyLoop(FitLoop):
#     def advance(self):
#         """Advance from one iteration to the next."""

#     def on_advance_end(self):
#         """Do something at the end of an iteration."""

#     def on_run_end(self):
#         """Do something when the loop ends."""


# class DefaultLoop(Loop):
#     def advance(self, batch, i):
#         loss = lightning_module.training_step(batch, i)
#         optimizer.zero_grad()
#         loss.backward()
#         optimizer.step()

#     def run(self, dataloader):
#         for i, batch in enumerate(dataloader):
#             self.advance(batch, i)


# from pytorch_lightning.loop import Loop


# class CustomLoop(Loop):
#     @property
#     def done(self) -> bool:
#         """Provide a condition to stop the loop."""

#     @property
#     def skip(self) -> bool:
#         """Determine whether to return immediately from the call to run()."""
#         pass

#     def reset(self) -> None:
#         """Resets the internal state of the loop at the beginning of each
#         call to run. Use when run is called multiple times."""
#         self.current_iteration = 0
#         self.outputs = []

#     def advance(self, *args, **kwargs):
#         """
#         Accepts all arguments passed to run.
#         Access your dataloader/s in whatever way you want.
#         Do your fancy optimization things.
#         Call the LightningModule methods at your leisure.
#         """
#         batch = next(iterator)
#         loss = self.trainer.lightning_module.training_step(batch, batch_idx)
#         ...

#     def run(self, *args, **kwargs):
#         if self.skip:
#             return self.on_skip()

#         self.reset()
#         self.on_run_start(*args, **kwargs)

#         while not self.done:
#             self.advance(*args, **kwargs)

#         output = self.on_run_end()
#         return output


# class ExampleCustomLoop(Loop):
#     @property
#     def done(self) -> bool:
#         return self.trainer.global_step >= self.trainer.max_steps

#     @property
#     def skip(self) -> bool:
#         return len(self.trainer.train_dataloader) == 0



from enum import Enum, auto
from typing import Protocol


class Stage(Enum):
    TRAIN = auto()
    TEST = auto()
    VAL = auto()


class StagePL(Enum):
    FIT = "fit"
    TEST = "test"
    PREDICT = "predict"


class ExperimentTracker(Protocol):
    def set_stage(self, stage: Stage):
        """Sets the current stage of the experiment."""

    def add_batch_metric(self, name: str, value: float, step: int):
        """Implements logging a batch-level metric."""

    def add_epoch_metric(self, name: str, value: float, step: int):
        """Implements logging a epoch-level metric."""



from dataclasses import dataclass
from typing import Optional

import pytorch_lightning as pl
import torch
import torch.nn as nn

from ..utils.config import _ABCDataClass
from ..utils.metacoding import CreditType
from ..utils.typescripts import Tensor
from .decoder import Decoder
from .encoder import Encoder
from .modelchoice import ModelChoice


@ModelChoice.register_arguments(call_name="pl_seq2seq")
@dataclass(unsafe_hash=True)
class Seq2SeqArguments(_ABCDataClass):
    src_vocab_size: int = 112
    tgt_vocab_size: int = 112
    num_encoder_layers: int = 3
    num_decoder_layers: int = 3
    emb_size: int = 512
    num_heads: int = 8
    dim_feedforward: int = 512
    dropout: float = 0.1
    weight_sharing: bool = False
    pad_idx: Optional[int] = None
    max_seq_len: Optional[int] = None


@ModelChoice.register_choice(
    call_name="pl_seq2seq",
    author="Peter Hartog",
    github_handle="PeterHartog",
    credit_type=CreditType.NONE,
)
class Seq2Seq(pl.LightningModule):
    def __init__(self, model_args: Seq2SeqArguments) -> None:
        super().__init__()

        self.save_hyperparameters()

        self.pad_idx = model_args.pad_idx

        self.encoder = Encoder(model_args=model_args)
        self.decoder = Decoder(model_args=model_args)
        self.fc_out = nn.Linear(model_args.emb_size, model_args.tgt_vocab_size)

    # (ignore mypy error using type: ignore)
    def forward(  # type: ignore
        self,
        src: Tensor,
        tgt: Tensor,
    ):
        src_mask = self.create_src_mask(src)
        tgt_mask = self.create_tgt_mask(tgt)
        src_padding_mask = (
            self.create_padding_mask(src, self.pad_idx)
            if self.pad_idx
            else None
        )
        tgt_padding_mask = (
            self.create_padding_mask(tgt, self.pad_idx)
            if self.pad_idx
            else None
        )

        memory = self.encoder(src, src_mask, src_padding_mask)

        x = self.decoder(
            tgt=tgt,
            memory=memory,
            tgt_mask=tgt_mask,
            memory_mask=None,
            tgt_padding_mask=tgt_padding_mask,
            memory_padding_mask=src_padding_mask,
        )
        out = self.fc_out(x)
        return out

    def generate_square_subsequent_mask(self, sz: int) -> Tensor:
        """Generates an upper-trianghular matrix of -inf with zeros on diag."""
        return torch.triu(
            torch.ones(sz, sz, device=self.device) * float("-inf"), diagonal=1
        )

    def create_src_mask(self, src: Tensor) -> Tensor:
        return torch.zeros(
            (src.shape[0], src.shape[0]), device=self.device
        ).type(torch.bool)

    def create_tgt_mask(self, tgt: Tensor) -> Tensor:
        return self.generate_square_subsequent_mask(tgt.shape[0])

    @staticmethod
    def create_padding_mask(
        seq: Tensor, pad_idx: int, eos_idx: int = 2
    ) -> Tensor:
        mask = torch.where((seq == pad_idx) | (seq == eos_idx), True, False)
        return mask.transpose(0, 1)



import torch
import torch.nn as nn
import torch.nn.functional as F
from typing import Any

from aidd_codebase.utils.metacoding import DictChoiceFactory
from aidd_codebase.utils.typescripts import Tensor


class LossChoice(DictChoiceFactory):
    pass


LossChoice.register_prebuilt_choice(
    call_name="cross_entropy", callable_cls=nn.CrossEntropyLoss
)
# Registers choices for loss functions of prebuilt loss functions
# LossChoice.register_prebuilt_choice(
#     call_name="focal_loss", callable_cls=nn.FocalLoss
# )
# LossChoice.register_prebuilt_choice(
#     call_name="kullback_leibler_div", callable_cls=nn.KLDivLoss
# )
# LossChoice.register_prebuilt_choice(
#     call_name="kullback_leibler_recon_div", callable_cls=nn.KLRecDivLoss
# )


class LogitLoss:
    def __init__(self, function: Any) -> None:
        self.criterium = function

    def __call__(self, logits: Tensor, tgt_out: Tensor):
        return self.criterium(
            logits.reshape(-1, logits.shape[-1]), tgt_out.reshape(-1)
        )  # .item()


# from https://www.kaggle.com/bigironsphere/loss-function-library-keras-pytorch
@LossChoice.register_choice("focal_loss")
class FocalLoss(nn.Module):
    def __init__(self):
        """Focal loss is to optimize for misclassified examples.
        This is often the case in diverse or imbalanced database.
        """
        super().__init__()

    def forward(
        self,
        inputs: Tensor,
        targets: Tensor,
        alpha: float = 0.8,
        gamma: int = 2,
    ):
        # comment out if your model contains a sigmoid or
        # equivalent activation layer
        inputs = F.sigmoid(inputs)

        # flatten label and prediction tensors
        inputs = inputs.view(-1)
        targets = targets.view(-1)

        # first compute binary cross-entropy
        BCE = F.binary_cross_entropy(inputs, targets, reduction="mean")
        BCE_EXP = torch.exp(-BCE)
        focal_loss = alpha * (1 - BCE_EXP) ** gamma * BCE

        return focal_loss


@LossChoice.register_choice("kullback_leibler_div")
class KLRecDivLoss(nn.Module):
    def __init__(self):
        """Kullback Leibler reconstruction divergence loss."""
        super().__init__()

    def forward(
        self,
        recon_x: Tensor,
        x: Tensor,
        mu,
        logvar,
    ) -> float:
        """see appendix B from VAE paper:
        Kingma and Welling. Auto-encoder Variational Bayes. ICLR, 2014
        https://arxiv.org/abs/1312.6114

        KLD = 0.5 + sum(1 + log(sigma^2) - mu^2 - sigma^2)
        """
        BCE = F.binary_cross_entropy(recon_x, x.view(-1, 784), reduction="sum")
        KLD = -0.5 + torch.sum(1 + logvar - mu.pow(2) - logvar.exp())

        return BCE + KLD


@LossChoice.register_choice("masked_soft_ce")
class MaskedSoftmaxCELoss(nn.CrossEntropyLoss):
    """The softmax cross-entropy loss with masks."""

    # `pred` shape: (`batch_size`, `num_steps`, `vocab_size`)
    # `label` shape: (`batch_size`, `num_steps`)
    # `valid_len` shape: (`batch_size`,)
    def forward(self, pred, label, valid_len, create_padding_mask):
        weights = torch.ones_like(label)
        mask = create_padding_mask(pred)
        weights = mask  # (weights, valid_len)  # create maskmask
        self.reduction = "none"
        unweighted_loss = super(MaskedSoftmaxCELoss, self).forward(
            pred.permute(0, 2, 1), label
        )
        weighted_loss = (unweighted_loss * weights).mean(dim=1)
        return weighted_loss



from typing import Any
import torch.nn as nn

from aidd_codebase.utils.metacoding import DictChoiceFactory
from aidd_codebase.utils.typescripts import Tensor


class LossChoice(DictChoiceFactory):
    pass


# Registers choices for loss functions of prebuilt loss functions
LossChoice.register_prebuilt_choice(call_name="cross_entropy", callable_cls=nn.CrossEntropyLoss)


@LossChoice.register_choice("logit_loss")
class LogitLoss:
    def __init__(self, function: Any) -> None:
        self.criterium = function

    def __call__(self, logits: Tensor, tgt_out: Tensor):
        return self.criterium(
            logits.reshape(-1, logits.shape[-1]), tgt_out.reshape(-1)
        )  # .item()  #(use loss.item to reserve memory)


@LossChoice.register_choice("focal_loss")
class FocalLoss:
    pass


@LossChoice.register_choice("ce_loss")
class CrossEntropyLoss:
    pass


@LossChoice.register_choice("poly_loss")
class PolyLoss:
    pass


@LossChoice.register_choice("kl_loss")
class KublitsLeibnerLoss:
    pass



import math

import torch.nn as nn

from aidd_codebase.utils.typescripts import Tensor


class TokenEmbedding(nn.Module):
    """Helper Module to convert input tensor to corresponding
    tensor of token embeddings."""

    def __init__(self, vocab_size: int, emb_size: int) -> None:
        super().__init__()
        self.embedding = nn.Embedding(vocab_size, emb_size)
        self.emb_size = emb_size

    def forward(self, tokens: Tensor) -> Tensor:
        return self.embedding(tokens.long()) * math.sqrt(self.emb_size)



from dataclasses import dataclass
from typing import List, Optional

import torch
from aidd_codebase.datamodules.datachoice import DataChoice
from aidd_codebase.utils.config import _ABCDataClass
from aidd_codebase.utils.metacoding import CreditType
from aidd_codebase.utils.tools import compose
from aidd_codebase.utils.typescripts import Tensor


@DataChoice.register_arguments(call_name="sequence_tokenizer")
@dataclass(unsafe_hash=True)
class TokenArguments(_ABCDataClass):
    # Define special symbols and indices
    pad_idx: int = 0  # Padding
    bos_idx: int = 1  # Beginning of Sequence
    eos_idx: int = 2  # End of Sequence
    unk_idx: int = 3  # Unknown Value
    msk_idx: Optional[int] = None  # Mask

    # Our vocabulary
    vocab: str = (
        " ^$?#%()+-./0123456789=@ABCDEFGHIKLMNOPRSTVXYZ[\\]abcdefgilmnoprstuy"
    )
    max_seq_len: int = 250


@DataChoice.register_choice(
    call_name="sequence_tokenizer",
    author="Peter Hartog",
    github_handle="PeterHartog",
    credit_type=CreditType.NONE,
)
class Tokenizer:
    def __init__(self, token_args: TokenArguments) -> None:
        self.vocab_size = len(token_args.vocab)
        self.char_to_ix = {ch: i for i, ch in enumerate(token_args.vocab)}
        self.ix_to_char = {i: ch for i, ch in enumerate(token_args.vocab)}
        self.pad_idx = token_args.pad_idx
        self.bos_idx = token_args.bos_idx
        self.eos_idx = token_args.eos_idx
        self.max_seq_len = token_args.max_seq_len
        self.max_tensor_len = self.max_seq_len + 2

        self.smile_prep = compose(self.smile_tokenizer, self.tensor_transform)
        self.smile_return = compose(
            self.reverse_tensor_transform, self.reverse_tokenizer
        )

    def smile_tokenizer(self, smile: str) -> List:
        return [self.char_to_ix[ch] for ch in smile]

    def reverse_tokenizer(self, tokens: List[int]) -> List:
        return [self.ix_to_char[token] for token in tokens]

    def tensor_transform(self, token_ids: List[int]) -> Tensor:
        """Function to add BOS/EOS, padding and create tensor for
        input sequence indices."""
        return torch.cat(
            (
                torch.tensor(
                    [self.bos_idx]
                    + token_ids
                    + [self.eos_idx]
                    + [self.pad_idx] * (self.max_seq_len - len(token_ids))
                ),
            )
        )

    def reverse_tensor_transform(self, tensor: Tensor):
        """Transforms a tensor back into a list of tokens."""
        tensor = tensor[
            torch.logical_and(
                torch.logical_and(
                    tensor != self.bos_idx, tensor != self.eos_idx
                ),
                tensor != self.pad_idx,
            )
        ]
        return tensor.tolist()



from typing import Callable, Optional

import torch.nn as nn

from ...models.modules.highway_units import (
    ConstantScalingUnit,
    ConvolutionShortcutUnit,
    DropoutShortcutUnit,
    HighwayUnit,
)
from aidd_codebase.utils.typescripts import Tensor


def unit_switch(unit: str) -> Callable:
    switch = {
        "highway": HighwayUnit,
        "constant_scaling": ConstantScalingUnit,
        "dropout_shortcut": DropoutShortcutUnit,
        "convolution_shortcut": ConvolutionShortcutUnit,
    }
    return switch.get(unit)


class HighwayRHCNN(nn.Module):
    def __init__(
        self,
        highway_type: str,
        n_layers: int,
        in_channels: int,
        out_channels: int,
        kernel_size: int,
        stride: int,
        dropout: Optional[float] = 0.1,
    ) -> None:
        super().__init__()
        assert highway_type in [
            "highway",
            "constant_scaling",
            "dropout_shortcut",
            "convolution_shortcut",
        ], (
            "unit type must be highway, constant_scaling, dropout_shortcut"
            + f"or convolution_shortcut, got {highway_type}"
        )
        unit = unit_switch(highway_type)
        if highway_type == "dropout_shortcut":
            self.highway = [
                unit(in_channels, out_channels, kernel_size, stride, dropout)
                for _ in range(n_layers)
            ]
        else:
            self.highway = [
                unit(in_channels, out_channels, kernel_size, stride)
                for _ in range(n_layers)
            ]

    def forward(self, x: Tensor) -> Tensor:
        for unit in self.highway:
            x = unit(x)
        return x



import warnings
from typing import Optional

import torch.nn as nn

from ...math_utils.cross_product import ScaledDotProductChoice
from aidd_codebase.utils.metacoding import DictChoiceFactory
from aidd_codebase.utils.typescripts import Tensor


class AttentionChoice(DictChoiceFactory):
    pass


AttentionChoice.register_prebuilt_choice(
    call_name="torch_multihead_attn", callable_cls=nn.MultiHeadAttention
)


# TODO build versions for skip_con|identical|batch_first


@AttentionChoice.register_choice("multihead_attn")
class MultiheadAttention(nn.Module):
    def __init__(
        self,
        qdim: int,
        kdim: int,
        vdim: int,
        num_heads: int,
        dropout: float = 0.0,
        bias: bool = True,
        add_bias_kv: bool = False,
        add_zero_attn: bool = False,
        batch_first: bool = False,
        dot_product_method: str = "matmul",
    ):
        super().__init__()
        assert (
            qdim % num_heads == 0
        ), "Embedding dimension must be divisible by the number of heads."

        self.qdim = qdim
        self.kdim = kdim
        self.vdim = vdim
        self.num_heads = num_heads
        self.head_dim = qdim // num_heads

        if qdim == kdim and qdim == vdim:
            self.qkv_proj = nn.Linear(qdim, 3 * qdim, bias)
        else:
            self.q_proj = nn.Linear(qdim, qdim, bias)
            self.k_proj = nn.Linear(kdim, qdim, bias)
            self.v_proj = nn.Linear(vdim, qdim, bias)

        self.o_proj = nn.Linear(qdim, qdim)

        self.dot_product = ScaledDotProductChoice.get_choice(
            dot_product_method
        )

        self.dropout = nn.Dropout(dropout)

    def forward(self, x, attn_mask=None):
        # query: [B,Q] (hidden state, decoder output, etc.)
        # keys: [T,B,K] (encoder outputs)
        # values: [T,B,V] (encoder outputs)
        # assume Q == K

        # not_batched = False
        # not_batch_first = False
        # if not batch_first_check(tensor=x):
        #     x = convert_to_batch_first(x)
        #     not_batch_first = True
        # if not is_batch(tensor=x):
        #     x = convert_unbatched(x)
        #     not_batched = True

        batch_size, seq_length, embed_dim = x.size()
        qkv = self.qkv_proj(x)

        # Separate Q, K, V from linear output
        qkv = qkv.reshape(
            batch_size, seq_length, self.num_heads, 3 * self.head_dim
        )
        qkv = qkv.permute(0, 2, 1, 3)  # [Batch, Head, SeqLen, Dims]
        q, k, v = qkv.chunk(3, dim=-1)

        # Determine value outputs
        values, attention = self.dot_product(
            q, k, v, attn_mask=attn_mask, dropout_p=self.dropout
        )
        values = values.permute(0, 2, 1, 3)  # [Batch, SeqLen, Head, Dims]
        values = values.reshape(batch_size, seq_length, embed_dim)
        o = self.o_proj(values)

        return o, attention


@AttentionChoice.register_choice("self_attn")
class SelfAttention(nn.Module):
    def __init__(self) -> None:
        super().__init__()

        self.attention = AttentionChoice.get_choice("self_attn")
        self.layernorm = nn.LayerNorm(input_dim)

    # layernorm then multihead

    def forward(self, x: Tensor, mask: Optional[Tensor]) -> Tensor:
        attn_out = self.self_attn(x, mask=mask)
        x = x + self.dropout(attn_out)
        x = self.norm1(x)

        residual = x
        x = self.attention(x, x, x)
        x += residual
        x = self.layernorm(x)
        return x

    pass


@AttentionChoice.register_choice("cross_attn")
class CrossAttention(nn.Module):
    # qv are from latent vector
    pass


@AttentionChoice.register_choice("geom_attn")
class GeometricAttention(nn.Module):
    pass


class Attention(nn.Module):
    def __init__(
        self,
        attention_type: str,
        embed_dim: int,
        kdim: int,
        vdim: int,
        num_heads: int,
        dropout: float = 0.0,
        bias: bool = True,
        add_bias_kv: bool = False,
        add_zero_attn: bool = False,
        batch_first: bool = False,
        dot_product_method: str = "matmul",
    ):
        super().__init__()
        AttentionChoice.validate_choice(attention_type)
        ScaledDotProductChoice.validate_choice(dot_product_method)

        if embed_dim != kdim and not attention_type == "cross_attn":
            warnings.warn("embed_dim must match kdim.")
            warnings.warn(f"Using cross_attn instead of {attention_type}.")
            attention = AttentionChoice.get_choice("cross_attn")
        else:
            attention = AttentionChoice.get_choice(attention_type)

        self.attention = attention(
            embed_dim=embed_dim,
            kdim=kdim,
            vdim=vdim,
            num_heads=num_heads,
            dropout=dropout,
            bias=bias,
            add_bias_kv=add_bias_kv,
            add_zero_attn=add_zero_attn,
            batch_first=batch_first,
        )

    def forward(
        self,
        q: Tensor,
        k: Tensor,
        v: Tensor,
        pad_mask: Optional[Tensor] = None,
        attn_mask: Optional[Tensor] = None,
    ):
        return self.attention(
            q, k, v, key_padding_mask=pad_mask, attn_mask=attn_mask
        )



from torch import optim

from aidd_codebase.utils.metacoding import DictChoiceFactory


class OptimizerChoice(DictChoiceFactory):
    pass


# Registers choices for loss functions of prebuilt loss functions
OptimizerChoice.register_prebuilt_choice(
    call_name="adam", callable_cls=optim.Adam
)



import torch.nn as nn

from aidd_codebase.utils.tools import DictChoiceFactory


class SwitchChoice(DictChoiceFactory):
    pass


@SwitchChoice.register_choice("excl_switch")
class ExclusiveSwitch(nn.Module):
    pass


@SwitchChoice.register_choice("moe_switch")
class MoESwitch(nn.Module):
    pass



from typing import Callable, Dict

from aidd_codebase.utils.config import _ABCDataClass
from aidd_codebase.utils.metacoding import Accreditation, DictChoiceFactory


class LoopChoice(DictChoiceFactory):
    pass


class IndependentLoopChoice(DictChoiceFactory):
    dict_choice: Dict[str, Callable] = {}
    choice_arguments: Dict[str, _ABCDataClass] = {}
    accreditation: Accreditation = Accreditation()



from typing import List, Tuple, Union

import numpy as np
import pandas as pd
from aidd_codebase.utils.typescripts import Tensor
from torch.utils.data import Dataset


class _AbsDataset(Dataset):
    """Dataset setup imported from the AstraZenica GitHub
    MolecularAI/Chemformer/blob/main/molbart/data/datasets.py"""

    def __len__(self) -> int:
        """Get the length of the dataset."""
        raise NotImplementedError()

    def __getitem__(
        self, idx: int
    ) -> Union[
        Tuple[np.ndarray, np.ndarray],
        Tuple[Tensor, Tensor],
        Tuple[pd.DataFrame, pd.DataFrame],
    ]:
        """Retrieve a tuple of source and target for a specific index."""
        raise NotImplementedError()


class DummyDataset(_AbsDataset):
    def __init__(self, data: List[zip]) -> None:
        self.data = [*data]

    def __len__(self) -> int:
        return len(self.data)

    def __getitem__(
        self, idx
    ) -> Union[
        Tuple[np.ndarray, np.ndarray],
        Tuple[Tensor, Tensor],
        Tuple[pd.DataFrame, pd.DataFrame],
    ]:
        inp, out = self.data[idx]
        return inp, out


class BasicDataset(Dataset):
    def __init__(
        self,
        data: pd.DataFrame,
    ) -> None:
        self.data = data
        self.input = (
            self.data.loc[:, (slice(None), "input")]
            .droplevel(level=1, axis=1)
            .squeeze()
            .to_numpy()
        )
        self.output = (
            self.data.loc[:, (slice(None), "output")]
            .droplevel(level=1, axis=1)
            .squeeze()
            .to_numpy()
        )

    def __len__(self) -> int:
        return len(self.data)

    def __getitem__(self, idx) -> Tuple[pd.DataFrame, pd.DataFrame]:
        return self.input[idx], self.output[idx]



import warnings
from functools import update_wrapper

import torch.nn as nn

from aidd_codebase.utils.typescripts import Tensor


class ResidualConnection:
    """Adds skip connections between layers."""

    module: nn.Module

    def __new__(cls, sequential: nn.Module):
        instance = super().__new__(cls)
        instance.module = sequential

        func = getattr(sequential, "forward")

        wrapped = ResidualConnection._residual_wrap(func)
        setattr(instance, "forward", wrapped)

        return instance

    def _residual_wrap(func):
        """
        Wraps *func* with additional code.
        """

        def wrapped(*args, **kwargs):
            if "residual" in kwargs:
                residual = kwargs.get("residual")

            output = func(*args, **kwargs)

            if type(output) == Tensor or (
                type(output) == tuple and type(output[0]) == Tensor
            ):
                output = (
                    residual + output
                    if type(output) != tuple
                    else residual + output[0]
                )
            else:
                warnings.warn("couldn't add residual connection")
            return output

        # Use "update_wrapper" to keep docstrings and other function metadata
        # intact
        update_wrapper(wrapped, func)

        return wrapped



import warnings
from typing import Dict, List, Optional

from aidd_codebase.utils.directories import Directories
from pytorch_lightning.loggers import (
    MLFlowLogger,
    TensorBoardLogger,
    WandbLogger,
)


class LoggerPL:
    pass


class PL_Loggers:
    loggers: Dict

    def __init__(
        self,
        directories: Directories,
        tensorboard: bool,
        wandb: bool,
        mlflow: bool,
    ) -> None:
        self.loggers = {}
        self.dirs = directories
        if tensorboard:
            self.init_tensorboard()
        if wandb:
            self.init_wandb()
        if mlflow:
            self.init_mlflow()

    def init_tensorboard(self) -> None:
        logger = TensorBoardLogger(
            save_dir=self.dirs.LOG_DIR, name=self.dirs.TIME
        )
        self.loggers["tensorboard"] = logger

    def init_wandb(self) -> None:
        logger = WandbLogger(
            save_dir=self.dirs.LOG_DIR,
            name=self.dirs.DATE,
            project=self.dirs.PROJECT,
        )
        self.loggers["wandb"] = logger

    def init_mlflow(self) -> None:
        logger = MLFlowLogger(
            experiment_name=self.dirs.PROJECT, tracking_uri="file:./ml-runs"
        )
        self.loggers["mlflow"] = logger

    def return_loggers(self) -> Optional[List]:
        if not self.loggers.values():
            warnings.warn("No loggers recorded")
            return None
        else:
            return list(self.loggers.values())



from dataclasses import dataclass
from typing import Optional

import pytorch_lightning as pl
import torch.nn as nn

from ..utils.config import _ABCDataClass
from ..utils.metacoding import CreditType
from ..utils.typescripts import Tensor
from .embedding.embedding import TokenEmbedding
from .embedding.positional import SequencePositionalEncoding
from .modelchoice import ModelChoice
from .modules.modules import WeightSharingEncoder


@ModelChoice.register_arguments(call_name="pl_encoder")
@dataclass(unsafe_hash=True)
class EncoderArguments(_ABCDataClass):
    src_vocab_size = 112
    num_encoder_layers = 3
    emb_size = 512
    num_heads = 8
    dim_feedforward = 512
    dropout: float = 0.1
    weight_sharing: bool = False
    max_seq_len: Optional[int] = None


@ModelChoice.register_choice(
    call_name="pl_encoder",
    author="Peter Hartog",
    github_handle="PeterHartog",
    credit_type=CreditType.NONE,
)
class Encoder(pl.LightningModule):
    def __init__(self, model_args: EncoderArguments) -> None:
        super().__init__()

        self.save_hyperparameters()

        self.src_vocab_size = model_args.src_vocab_size
        self.num_encoder_layers = model_args.num_encoder_layers
        self.emb_size = model_args.emb_size
        self.num_heads = model_args.num_heads
        self.dim_feedforward = model_args.dim_feedforward
        self.dropout = model_args.dropout

        self.positional_encoding = SequencePositionalEncoding(
            emb_size=model_args.emb_size,
            dropout=model_args.dropout,
            maxlen=model_args.max_seq_len,
        )
        self.src_tok_emb = TokenEmbedding(
            vocab_size=model_args.src_vocab_size, emb_size=model_args.emb_size
        )

        norm = nn.LayerNorm(model_args.emb_size)

        encoder_layer = nn.TransformerEncoderLayer(
            d_model=model_args.emb_size,
            nhead=model_args.num_heads,
            dim_feedforward=model_args.dim_feedforward,
            dropout=model_args.dropout,
            activation="relu",
        )

        if model_args.weight_sharing:
            self.encoder = WeightSharingEncoder(
                encoder_layer=encoder_layer,
                num_layers=model_args.num_encoder_layers,
                norm=norm,
            )
        else:
            self.encoder = nn.TransformerEncoder(
                encoder_layer=encoder_layer,
                num_layers=model_args.num_encoder_layers,
                norm=norm,
            )

    # (ignore mypy error using type: ignore)
    def forward(  # type: ignore
        self,
        src: Tensor,
        mask: Optional[Tensor] = None,
        src_padding_mask: Optional[Tensor] = None,
    ) -> Tensor:
        src_embedded = self.positional_encoding(self.src_tok_emb(src))
        return self.encoder(src_embedded, mask, src_padding_mask)



import math
from typing import Optional

import torch
import torch.nn.functional as F
from aidd_codebase.utils.metacoding import DictChoiceFactory
from aidd_codebase.utils.typescripts import Tensor


class ScaledDotProductChoice(DictChoiceFactory):
    pass


@ScaledDotProductChoice.register_choice("matmul")
def scaled_dot_product_matmul(
    q: Tensor,
    k: Tensor,
    v: Tensor,
    attn_mask: Optional[Tensor] = None,
    dropout_p: float = 0.0,
) -> Tensor:
    # attn = (B, Nt, E) x (B, E, Ns) -> (B, Nt, Ns)
    # output = (B, Nt, E) x (B, E, Ns) -> (B, Nt, E)
    # q = [B, Nt, E]; k = [B, Ns, E]; v = [B, Ns, E]
    # B: batch; E: embedding; Nt: target size; Ns: source size
    batch_size, target_seq_length, embed_dim = q.shape
    q = q / math.sqrt(embed_dim)

    attn = torch.matmul(q, k.transpose(-2, -1))
    if attn_mask is not None:
        attn = attn.masked_fill(attn_mask == 0, -1e9)

    attn = F.softmax(attn, dim=-1)

    if dropout_p > 0.0:
        attn = F.dropout(attn, p=dropout_p)

    output = torch.matmul(attn, v.transpose(-2, -1))
    return output, attn


@ScaledDotProductChoice.register_choice("einsum")
def scaled_dot_product_einsum(
    q: Tensor,
    k: Tensor,
    v: Tensor,
    attn_mask: Optional[Tensor] = None,
    dropout_p: float = 0.0,
) -> Tensor:
    # attn = (B, Nt, E) x (B, E, Ns) -> (B, Nt, Ns)
    # output = (B, Nt, E) x (B, E, Ns) -> (B, Nt, E)
    # q = [B, Nt, E]; k = [B, Ns, E]; v = [B, Ns, E]
    # B: batch; E: embedding; Nt: target size; Ns: source size
    batch_size, target_seq_length, embed_dim = q.shape
    q = q / math.sqrt(embed_dim)

    attn = torch.einsum("BtE,BsE->Bts", q, k)
    if attn_mask is not None:
        attn = attn.masked_fill(attn_mask == 0, -1e9)

    attn = F.softmax(attn, dim=-1)

    if dropout_p > 0.0:
        attn = F.dropout(attn, p=dropout_p)

    output = torch.einsum("Bts,BsE->BtE", attn, v)
    return output, attn


@ScaledDotProductChoice.register_choice("bmm")
def scaled_dot_product_bmm(
    q: Tensor,
    k: Tensor,
    v: Tensor,
    attn_mask: Optional[Tensor] = None,
    dropout_p: float = 0.0,
) -> Tensor:
    # attn = (B, Nt, E) x (B, E, Ns) -> (B, Nt, Ns)
    # output = (B, Nt, E) x (B, E, Ns) -> (B, Nt, E)
    # q = [B, Nt, E]; k = [B, Ns, E]; v = [B, Ns, E]
    # B: batch; E: embedding; Nt: target size; Ns: source size
    batch_size, target_seq_length, embed_dim = q.shape
    q = q / math.sqrt(embed_dim)

    if attn_mask is not None:
        attn = torch.baddbmm(attn_mask, q, k.transpose(-2, -1))
    else:
        attn = torch.bmm(q, k.transpose(-2, -1))

    attn = F.softmax(attn, dim=-1)
    if dropout_p > 0.0:
        attn = F.dropout(attn, p=dropout_p)

    output = torch.bmm(attn, v)
    return output, attn



import torch.nn as nn

from .attention import AttentionChoice


class EncoderBlock(nn.Module):
    def __init__(
        self,
        input_dim,
        num_heads,
        dim_feedforward,
        dropout: float = 0.0,
        attention_type: str = "multihead_attn",
    ) -> None:
        """
        Inputs:
            input_dim - Dimensionality of the input
            num_heads - Number of heads to use in the attention block
            dim_feedforward - Dimensionality of the hidden layer in the MLP
            dropout - Dropout probability to use in the dropout layers
        """
        super().__init__()

        # Attention layer
        self.attention = AttentionChoice.get_choice(attention_type)
        self.attention = self.attention(input_dim, input_dim, num_heads)

        # Two-layer MLP
        self.linear_net = nn.Sequential(
            nn.Linear(input_dim, dim_feedforward),
            nn.Dropout(dropout),
            nn.ReLU(inplace=True),
            nn.Linear(dim_feedforward, input_dim),
        )

        # Layers to apply in between the main layers
        self.norm1 = nn.LayerNorm(input_dim)
        self.norm2 = nn.LayerNorm(input_dim)
        self.dropout = nn.Dropout(dropout)

    def forward(self, x, mask=None):
        # Attention part
        attn_out = self.self_attn(x, mask=mask)
        x = x + self.dropout(attn_out)
        x = self.norm1(x)

        # MLP part
        linear_out = self.linear_net(x)
        x = x + self.dropout(linear_out)
        x = self.norm2(x)

        return x



import torch.nn as nn

from aidd_codebase.utils.metacoding import DictChoiceFactory


class AdapterChoice(DictChoiceFactory):
    pass


# class AdapterABC(nn.Module):
#     def __init__(self) -> None:
#         super().__init__()

#     def forward(
#         self, x: Tensor, mask: Tensor = None
#     ) -> Union[Tensor, Tuple[Tensor, Tensor]]:
#         raise NotImplementedError


class InputAdapter(nn.Module):
    def __init__(self) -> None:
        super().__init__()

    def forward(self):
        raise NotImplementedError

    @property
    def positional_encoding(self):
        raise NotImplementedError


class SequenceAdapter(InputAdapter):
    # TODO
    def __init__(self) -> None:
        super().__init__()

    def positional_encoding(self):
        return super().positional_encoding()  # identical? think not

    def forward(self):
        pass


class ImageAdapter(InputAdapter):
    # TODO
    def __init__(self) -> None:
        super().__init__()

    def positional_encoding(self):
        return super().positional_encoding()

    def forward(self):
        pass


class GraphAdapter(InputAdapter):
    # TODO
    def __init__(self) -> None:
        super().__init__()

    def positional_encoding(self):
        return super().positional_encoding()

    def forward(self):
        pass



from pytorch_lightning.callbacks import Callback

from .loading_saving import save_model


class CustomSavingCallback(Callback):
    def __init__(
        self, model_save_path: str, save_weights: bool = True
    ) -> None:
        super().__init__()
        self.model_save_path = model_save_path
        self.save_weights = save_weights

    def on_validation_epoch_end(self, trainer, *args, **kwargs) -> None:
        save_model(
            model=trainer.model.model,
            name=f"model-epoch={trainer.current_epoch}",
            path=self.model_save_path,
            save_weights=self.save_weights,
        )



import torch
import torch.nn as nn


class ParameterInitialization:
    def __init__(self, method: str = "xavier") -> None:
        self.initialize_method = method
        if method not in ["xavier", "kaiming"]:
            raise ValueError(f"initialization {method} is not recognized.")

    def initialize_model(self, model: nn.Module) -> nn.Module:
        if self.initialize_method == "xavier":
            return self.xavier_initialization(model)
        elif self.initialize_method == "kaiming":
            return self.kaiming_initialization(model)

    @staticmethod
    def xavier_initialization(model: nn.Module) -> nn.Module:
        for p in model.parameters():
            if p.dim() > 1:
                torch.nn.init.xavier_uniform_(p)
        return model

    @staticmethod
    def kaiming_initialization(model: nn.Module) -> nn.Module:
        for p in model.parameters():
            if p.dim() > 1:
                torch.nn.init.kaiming_normal_(p, mode="fan_out")
        return model



from dataclasses import dataclass
from typing import List, Optional

import torch


@dataclass
class Device:
    device: str
    accelerator: Optional[str] = None
    precision: int = 32
    multi_gpu: bool = False
    gpus: Optional[List[int]] = None

    def __post_init__(self) -> None:
        if self.device == "cpu" or not torch.cuda.is_available():
            self.torch_device = torch.device("cpu")
            self.accelerator = "cpu"
            self.precision = 32
            self.gpus = None
        elif self.device.startswith("cuda") and torch.cuda.is_available():
            print("we are here")
            self.torch_device = torch.cuda.device("cuda")
            self.accelerator = "gpu"

            if torch.cuda.device_count() == 1:
                self.multi_gpu = False

            if ":" in self.device:
                self.gpus = [
                    int(i) for i in self.device.split(":")[-1].split(",")
                ]
                if not self.multi_gpu:
                    self.gpus = [self.gpus[0]]
            else:
                self.gpus = [i for i in range(torch.cuda.current_device())]
                if not self.multi_gpu:
                    self.gpus = [0]

    def display(self) -> None:
        print(f"\nGPU is available: {torch.cuda.is_available()}")
        return
        if torch.cuda.is_available() and self.device.startswith("cuda"):
            with self.torch_device:
                print(
                    f"Using device {str(torch.cuda.current_device())}"
                    + f"/{str(torch.cuda.device_count())}, "
                    + f"name: {str(torch.cuda.get_device_name(0))}."
                )



from typing import List, Optional, Tuple

import torch
from aidd_codebase.utils.typescripts import Tensor
from torch.nn.utils.rnn import pad_sequence


class Collate:
    def __init__(self, pad_idx: Optional[int]) -> None:
        self.pad_idx = pad_idx

    def seq_collate_fn(
        self, batch: Tuple[List[Tensor], List[Tensor]]
    ) -> Tuple[Tensor, Tensor]:
        """Function to collate data samples into batch tensors"""
        src_batch, tgt_batch = [], []
        for src, tgt in batch:
            src_sample = src
            tgt_sample = tgt

            src_batch.append(src_sample)
            tgt_batch.append(tgt_sample)

        src_batch = pad_sequence(src_batch, padding_value=self.pad_idx)
        tgt_batch = pad_sequence(tgt_batch, padding_value=self.pad_idx)
        return src_batch, tgt_batch

    def simple_collate_fn(
        self, batch: Tuple[List[Tensor], List[Tensor]]
    ) -> Tuple[Tensor, Tensor]:
        """Stacks input tensors into source and target tensors.
        Assumes all tensors of equal length.
        """
        src_batch, tgt_batch = zip(*batch)
        return torch.t(torch.stack(src_batch)), torch.t(torch.stack(tgt_batch))



from os import listdir
from os.path import isfile, join
import pathlib
from typing import Any, Dict, List, Optional

import pandas as pd
import pytorch_lightning as pl
import torch
from dataclasses import dataclass, field
from aidd_codebase.data_utils.augmentation import Enumerator
from aidd_codebase.data_utils.collate import Collate
from aidd_codebase.utils.config import _ABCDataClass
from aidd_codebase.datamodules.datachoice import DataChoice
from aidd_codebase.utils.metacoding import CreditType
from aidd_codebase.data_utils.dataprocessors import _AbsProcessor, DataType, ReturnOptions, SmilesProcessor
from aidd_codebase.data_utils.datasets import BasicDataset
from aidd_codebase.utils.directories import validate_or_create_dir
from aidd_codebase.utils.typescripts import Tensor
from torch.utils.data import DataLoader


@DataChoice.register_arguments(call_name="retrosynthesis_pavel")
@dataclass(unsafe_hash=True)
class DataArguments(_ABCDataClass):
    seed: int = 1234
    batch_size: int = 128
    num_workers: int = 8
    persistent_workers: bool = True

    override_prepared_data: bool = False
    prepared_data_dir: Optional[str] = f"{pathlib.Path(__file__).parent.resolve()}/data/saved"

    #partitions: Dict[str, float] = field(default_factory=lambda: {"train": 0.8, "val": 0.1, "test": 0.1})

    remove_missing: bool = True
    remove_duplicates: bool = False
    canonicalization: bool = True
    enumeration: int = 0  # 10
    enumeration_oversample: int = 0  # 15


@DataChoice.register_choice(call_name="retrosynthesis_pavel", author="Peter Hartog", github_handle="PeterHartog",
                            credit_type=CreditType.NONE)
class SmilesDataModule(pl.LightningDataModule):
    def __init__(
            self,
            tokenizer: Any,
            data_args: DataArguments,
    ):
        super().__init__()

        self.datasplit: Dict[str, pd.DataFrame] = {}

        self.prepared_data_dir = validate_or_create_dir(data_args.prepared_data_dir)
        self.override_prepared_data = data_args.override_prepared_data

        self.seed = data_args.seed
        self.partitions = {"train": 0.8, "val": 0.1, "test": 0.1} #data_args.partitions
        self.batch_size = data_args.batch_size
        self.num_workers = data_args.num_workers
        self.persistent_workers = data_args.persistent_workers
        self.collate_fn = Collate(tokenizer.pad_idx).simple_collate_fn

        enumerator = Enumerator(
            enumerations=data_args.enumeration,
            seed=data_args.seed,
            oversample=data_args.enumeration_oversample,
            max_len=tokenizer.max_seq_len,
            keep_original=data_args.canonicalization,
        )
        if data_args.enumeration > 0:
            augmentations = [lambda x: enumerator.dataframe_enumeration(x)]
        else:
            augmentations = None

        self.input_processor = SmilesProcessor(
            type=DataType.SEQUENCE,
            return_type=ReturnOptions.CLEANED
            if data_args.canonicalization
            else ReturnOptions.Original,
            tokenizer=tokenizer,
            remove_duplicates=data_args.remove_duplicates,
            remove_missing=data_args.remove_missing,
            constrains=[lambda x: x.applymap(len) <= tokenizer.max_seq_len],
            augmentations=augmentations,
        )
        self.output_processor = SmilesProcessor(
            type=DataType.SEQUENCE,
            return_type=ReturnOptions.Original,
            tokenizer=tokenizer,
            remove_duplicates=data_args.remove_duplicates,
            remove_missing=data_args.remove_missing,
            constrains=[lambda x: x.applymap(len) <= tokenizer.max_seq_len],
            augmentations=None,
        )

    def process_data(
            self, data: pd.DataFrame, data_processor: _AbsProcessor
    ) -> pd.DataFrame:
        data_processor.set_data(data)
        data_processor.inspect_data()
        data_processor.clean_data()
        data_processor.clean_report()
        return data_processor.return_data()

    def split_data(
            self,
            input: pd.DataFrame,
            output: pd.DataFrame,
            partitions: Dict[str, float],
    ) -> None:
        assert len(input) == len(
            output
        ), "The input length must match the output length."
        assert (
                sum(partitions.values()) <= 1
        ), "Combined partitions should be less than or equal to 1"

        df = self.combine_input_output(input, output)
        corrected_fractions = self.correct_fractions(list(partitions.values()))

        datasets: List = []
        sampled_idxs: List = []
        for frac in corrected_fractions:
            fraction_dataset = df.drop(sampled_idxs).sample(
                frac=frac, random_state=self.seed
            )
            datasets.append(fraction_dataset)
            sampled_idxs.extend(fraction_dataset.index)

        self.datasplit.update(zip(list(partitions.keys()), datasets))

        print(
            ", ".join(
                [f"{k.upper()}: {len(v)}" for k, v in self.datasplit.items()]
            )
        )

    def prepare_data(self, **kwargs) -> None:
        print("Starting input data download...")
        url = 'https://raw.githubusercontent.com/bigchem/retrosynthesis/master/data/retrosynthesis-all.smi'
        reactions = pd.read_csv(url, header=None, names=["reaction"])
        reactions = reactions.applymap(lambda x: x.split(">>"))
        reactions = pd.DataFrame(reactions["reaction"].to_list(), columns=['product', 'reactants'])

        self.input = reactions.loc[:, ["product"]]
        print(self.input.head())
        self.input = self.process_data(self.input, self.input_processor)

        self.target = reactions.loc[:, ["reactants"]]
        print(self.target.head())
        self.output = self.process_data(self.target, self.output_processor)

        self.split_data(
            self.input, self.output, partitions=self.partitions
        )

    def save_prepared_data(self, data: pd.DataFrame, stage: str) -> None:
        input = data.loc[:, (slice(None), "input")].droplevel(level=1, axis=1)
        output = data.loc[:, (slice(None), "output")].droplevel(
            level=1, axis=1
        )
        input.to_csv(
            f"{self.prepared_data_dir}/{stage}_input.csv", index=False
        )
        output.to_csv(
            f"{self.prepared_data_dir}/{stage}_output.csv", index=False
        )

    def load_prepared_data(self, stage: str) -> pd.DataFrame:
        print(f"Loading {stage}...")
        input = pd.read_csv(f"{self.prepared_data_dir}/{stage}_input.csv")
        output = pd.read_csv(f"{self.prepared_data_dir}/{stage}_output.csv")
        return self.combine_input_output(input, output)

    def transform_data(
            self,
            data: pd.DataFrame,
    ) -> pd.DataFrame:
        input = data.loc[:, (slice(None), "input")].droplevel(level=1, axis=1)
        output = data.loc[:, (slice(None), "output")].droplevel(
            level=1, axis=1
        )
        input = self.input_processor.transform_data(input)
        output = self.output_processor.transform_data(output)
        return self.combine_input_output(input, output)

    def augment_data(self, data: pd.DataFrame) -> pd.DataFrame:
        input = data.loc[:, (slice(None), "input")].droplevel(level=1, axis=1)
        output = data.loc[:, (slice(None), "output")].droplevel(
            level=1, axis=1
        )
        input = self.input_processor.augment_data(input)
        output = self.output_processor.augment_data(output)
        return self.combine_input_output(input, output)

    def setup_data(self, data: pd.DataFrame, stage: str) -> pd.DataFrame:
        print(f"Setting up {stage} stage.")
        print("Augmenting data...")
        data = self.augment_data(data)
        self.save_prepared_data(data, stage)
        return data.reset_index(drop=True)

    def get_data(self, partition: str) -> pd.DataFrame:
        if (not self.override_prepared_data) and self.check_prepared_files(
                dir=self.prepared_data_dir, stages=[partition]
        ):
            partition_data = self.load_prepared_data(partition)
        else:
            if partition not in self.datasplit.keys():
                raise ValueError(f"Partition {partition} not in partitions.")
            partition_data = self.setup_data(
                self.datasplit[partition], partition
            )
        partition_data = self.transform_data(partition_data)
        return partition_data

    def setup(self, stage: Optional[str] = None) -> None:
        print("Starting Data Module Setup...")

        if stage in ("fit", None):
            self.train = self.get_data(partition="train")
            self.val = self.get_data(partition="val")

        if stage in ("test", None):
            self.test = self.get_data(partition="test")

        print("Finished Data Module!")

    def train_dataloader(self):
        return DataLoader(
            BasicDataset(self.train),
            batch_size=self.batch_size,
            shuffle=True,
            collate_fn=self.collate_fn,
            num_workers=self.num_workers if torch.cuda.is_available() else 0,
            persistent_workers=self.persistent_workers
            if torch.cuda.is_available()
            else False,
            pin_memory=torch.cuda.is_available(),
        )

    def val_dataloader(self):
        return DataLoader(
            BasicDataset(self.val),
            batch_size=self.batch_size,
            shuffle=False,
            collate_fn=self.collate_fn,
            num_workers=self.num_workers if torch.cuda.is_available() else 0,
            persistent_workers=self.persistent_workers
            if torch.cuda.is_available()
            else False,
            pin_memory=torch.cuda.is_available(),
        )

    def test_dataloader(self):
        return DataLoader(
            BasicDataset(self.test),
            batch_size=self.batch_size,
            shuffle=False,
            collate_fn=self.collate_fn,
            num_workers=self.num_workers if torch.cuda.is_available() else 0,
            persistent_workers=self.persistent_workers
            if torch.cuda.is_available()
            else False,
            pin_memory=torch.cuda.is_available(),
        )

    @staticmethod
    def combine_input_output(
            input: pd.DataFrame, output: pd.DataFrame
    ) -> pd.DataFrame:
        input.columns = pd.MultiIndex.from_product([input.columns, ["input"]])
        output.columns = pd.MultiIndex.from_product(
            [output.columns, ["output"]]
        )
        return pd.merge(input, output, left_index=True, right_index=True)

    @staticmethod
    def check_prepared_files(dir: str, stages: List[str]) -> bool:
        files = [f for f in listdir(dir) if isfile(join(dir, f))]
        opties = ["input.csv", "output.csv"]
        required_files = ["_".join([s, o]) for s in stages for o in opties]
        return all(file in files for file in required_files)

    @staticmethod
    def correct_fractions(fractions: List[float]) -> List[float]:
        """Corrects the fractions to the fraction
        of the data that is left after previous fractions are removed.
        """
        corrected_fractions = [
            frac / (1.0 - sum(fractions[:idx]))
            for idx, frac in enumerate(fractions)
        ]
        # to prevent weird error of 0.1/0.1 > 1.0
        corrected_fractions = [
            frac if frac <= 1.0 else 1.0 for frac in corrected_fractions
        ]
        return corrected_fractions



from typing import Any, Callable, Dict, List, Optional, Union

import pytorch_lightning as pl
import torch
from aidd_codebase.models.metrics.metrics import Metric
from aidd_codebase.models.optimizers.scheduling import Scheduler
from aidd_codebase.utils.typescripts import Tensor

from .loggers import LoggerPL


class ModelFramework(pl.LightningModule):
    def __init__(
        self,
        loss,
        optimizer,
        metrics: Optional[List[Metric]] = None,
        scheduler: Optional[Scheduler] = None,
        pl_loggers: Optional[List[LoggerPL]] = None,
    ) -> None:
        super().__init__()

        # saving parameters
        self.save_hyperparameters()

        # setting parameters
        self.loss = loss
        self.metrics = metrics
        self.optimizer = optimizer
        self.scheduler = scheduler
        self.pl_loggers = pl_loggers

        self.metrics_list: Dict[str, Metric] = {}

    def set_model(self, model: pl.LightningModule) -> None:
        self.model = model

    def set_loop(self, loop: Callable) -> None:
        self.training_step_imported = loop
        self.validation_step_imported = loop
        self.test_step_imported = loop

    def training_step(self, batch: Any, batch_idx: int):
        return self.training_step_imported(
            self, batch, batch_idx, stage="train"
        )

    def validation_step(self, batch: Any, batch_idx: int):
        return self.validation_step_imported(
            self, batch, batch_idx, stage="validation"
        )

    def test_step(self, batch: Any, batch_idx: int):
        return self.test_step_imported(self, batch, batch_idx, stage="test")

    def forward(self, *args, **kwargs):
        return self.model(*args, **kwargs)

    def optimizer_step(self, *args, **kwargs):
        super().optimizer_step(*args, **kwargs)
        if self.scheduler:
            self.scheduler.step()

    def configure_optimizers(self):
        optimizer = self.optimizer

        if self.scheduler:
            self.scheduler = self.scheduler.set_optimizer(optimizer)

        return optimizer

    def _init_metrics(self, stage: str) -> None:
        loss_key = self.get_loss_key(stage)
        self.metrics_list[stage] = {
            k: Metric()
            for k in list(
                [(loss_key, *self.metrics) if self.metrics else loss_key]
            )
        }

    def _log_dict(
        self, dict: Dict, batch_size: Optional[Tensor] = None
    ) -> None:
        """Logs everything in a dict."""
        for key, value in dict.items():
            self.log(key, value, batch_size=batch_size)

    def _log_metrics(
        self,
        name: str,
        metric: Metric,
        batch_size: Optional[Tensor] = None,
    ) -> None:
        self.log(
            name,
            metric.values[-1],
            batch_size=batch_size,
        )

    def _log_epoch_metrics(self, metrics: List[Metric]) -> None:
        if self.pl_loggers:
            for logger in self.pl_loggers:
                logger.log_scalar(metrics=metrics, epoch=self.current_epoch)

    def on_train_start(self) -> None:
        self._init_metrics(stage="train")

    def on_training_epoch_start(self) -> None:
        """Called before every training epoch."""
        self._init_metrics(stage="train")

    def on_validation_epoch_start(self) -> None:
        """Called before every validation epoch."""
        self._init_metrics(stage="validation")

    def on_test_epoch_start(self) -> None:
        """Called before every test epoch."""
        self._init_metrics(stage="test")

    def on_training_batch_end(
        self,
        outputs: Any,
        batch: Any,
        batch_idx: int,
        dataloader_idx: int,
    ) -> None:
        x, _ = batch
        batch_size = x.shape[1]
        print(self.metrics_list)
        print(self.metrics_list["train"])
        for name, metric in self.metrics_list["train"].items():
            self._log_metrics(
                name,
                metric,
                torch.tensor(batch_size, device=self.device),
            )

    def on_validation_batch_end(
        self,
        outputs: Union[Tensor, Dict[str, Any], None],
        batch: Any,
        batch_idx: int,
        dataloader_idx: int,
    ) -> None:
        x, _ = batch
        batch_size = x.shape[1]
        for name, metric in self.metrics_list["validation"].items():
            self._log_metrics(
                name,
                metric,
                torch.tensor(batch_size, device=self.device),
            )

    def on_test_batch_end(
        self,
        outputs: Union[Tensor, Dict[str, Any], None],
        batch: Any,
        batch_idx: int,
        dataloader_idx: int,
    ) -> None:
        x, _ = batch
        batch_size = x.shape[1]
        for name, metric in self.metrics_list["test"].items():
            self._log_metrics(
                name,
                metric,
                torch.tensor(batch_size, device=self.device),
            )

    def training_epoch_end(self, outputs: List[Union[Tensor, Dict[str, Any]]]):
        """This function is called after every epoch"""
        self._log_epoch_metrics(self.metrics_list["train"])

    def validation_epoch_end(
        self, outputs: List[Union[Tensor, Dict[str, Any]]]
    ):
        """This function is called after every epoch"""
        self._log_epoch_metrics(self.metrics_list["validation"])

    def test_epoch_end(self, outputs: List[Union[Tensor, Dict[str, Any]]]):
        """This function is called after every epoch"""
        self._log_epoch_metrics(self.metrics_list["test"])

    @staticmethod
    def get_loss_key(stage: str) -> str:
        if stage == "train":
            loss_key = "loss"
        elif stage == "validation":
            loss_key = "val_loss"
        else:
            loss_key = "test_loss"

        return loss_key



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['aidd_codebase',
 'aidd_codebase.data_utils',
 'aidd_codebase.datamodules',
 'aidd_codebase.framework',
 'aidd_codebase.math_utils',
 'aidd_codebase.models',
 'aidd_codebase.models.embedding',
 'aidd_codebase.models.metrics',
 'aidd_codebase.models.modules',
 'aidd_codebase.models.optimizers',
 'aidd_codebase.utils']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'aidd-codebase',
    'version': '0.1.6',
    'description': 'High-level codebase for deep learning development in drug discovery.',
    'long_description': '# AIDD Codebase\n\n![PyPI](https://img.shields.io/pypi/v/aidd-codebase)\n![PyPI](https://img.shields.io/pypi/pyversions/aidd-codebase)\n![PyPI](https://img.shields.io/github/license/AIDD-ESR1/AIDD-codebase)\n\nA high-level codebase for deep learning development in drug discovery applications using PyTorch-Lightning.\n\n## Dependencies\n\nThe codebase requires the following additional dependencies\n- CUDA >= 11.4\n- PyTorch >= 1.9\n- Pytorch-Lightning >= 1.5 \n- RDKit \n- Optionally supports: tensorboard and/or wandb\n\n\n## Installation\n\nThe codebase can be installed from PyPI using `pip`, or your package manager of choice, with\n\n```\npip install aidd-codebase\n```\n\n## Usage\n\n1. __*Configuration*__: The coding framework has a number of argument dataclasses in the file *arguments.py*. This file contains all standard arguments for each of the models. Because they are dataclasses, you can easily adapt them to your own needs. \n<br> \nDoes your Seq2Seq adaptation need an extra argument? Import the Seq2SeqArguments from arguments.py, create your own dataclass which inherits it and add your extra argument. <br> <br>\n*It is important to note that the order of supplying arguments to a script goes as follows:* <br>\n- --flags override config.yaml <br>\n- config.yaml overrides default values in arguments.py <br>\n- default values from arguments.py are used when no other values are supplied<br>\nAt the end, it stores all arguments in config.yaml\n<br><br>\n\n2. __*Use*__: The coding framework has four main parts: <br>\n- utils\n- data_utils\n- models\n- interpretation\n\nThese parts should be used \n&nbsp; \n\n3. __*File Setup*__: The setup of the files in the system is best used as followed:<br>\ncoding_framework<br> \n|-- ..<br> \nESR X<br> \n|-- project 1<br> \n  |-- data<br> \n    |-- ..<br> \n  |-- Arguments.py<br> \n  |-- config.yaml<br> \n  |-- main.py<br>\n  |-- datamodule.py<br>\n  |-- pl_framework.py<br>\n\n## Contributors\n\nAll fellows of the AIDD consortium have contributed to the packaged.\n\n## Code of Conduct\n\nEveryone interacting in the codebase, issue trackers, chat rooms, and mailing lists is expected to follow the [PyPA Code of Conduct](https://www.pypa.io/en/latest/code-of-conduct/).\n\n ',
    'author': 'Peter Hartog',
    'author_email': 'peter.hartog@hotmail.nl',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/AIDD-ESR1/AIDD-codebase',
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import math
from typing import Optional

import torch
import torch.nn as nn

from aidd_codebase.utils.metacoding import DictChoiceFactory
from aidd_codebase.utils.typescripts import Tensor

# Positional Encoding Options
# adding vs concatonating | static vs learned |
# relative (clipping at k) vs absolute | Type: Image vs Graph vs Sequence


class PositionalChoice(DictChoiceFactory):
    pass


class LearnedChoice(DictChoiceFactory):
    pass


class CombinationChoice(DictChoiceFactory):
    pass


class PositionalABC(nn.Module):
    pass


@PositionalChoice.register_choice("relative")
class RelativeEncoding(PositionalABC):
    def __init__(self) -> None:
        super().__init__()


@PositionalChoice.register_choice("relative_clipped")
class RelativeClippedEncoding(PositionalABC):
    def __init__(self) -> None:
        super().__init__()


@PositionalChoice.register_choice("absolute")
class AbsoluteEncoding(PositionalABC):
    def __init__(self) -> None:
        super().__init__()


@LearnedChoice.register_choice("static")
class StaticEncoding(nn.Module):
    def __init__(self) -> None:
        super().__init__()


@LearnedChoice.register_choice("learned")
class LearnedEncoding(nn.Module):
    def __init__(self) -> None:
        super().__init__()


class PositionalEncoding(nn.Module):
    def __init__(
        self, positional: str, combination: str, learned_option: str
    ) -> None:
        super().__init__()
        self.pos_encoding = nn.Sequential(
            PositionalChoice.get_choice(positional),
            CombinationChoice.get_choice(combination),
            LearnedChoice.get_choice(learned_option),
        )


# class AttentionABC(nn.Module):
#     def __init__(self) -> None:
#         super().__init__()

#     def forward(
#         self, x: Tensor, mask: Tensor = None
#     ) -> Union[Tensor, Tuple[Tensor, Tensor]]:
#         raise NotImplementedError


class SequencePositionalEncoding(nn.Module):
    """Helper Module that adds positional encoding to sequence embedding."""

    def __init__(
        self, emb_size: int, dropout: float, maxlen: Optional[int] = None
    ) -> None:
        super().__init__()
        if not maxlen:
            maxlen = 5000

        den = torch.exp(
            -torch.arange(0, emb_size, 2) * math.log(10000) / emb_size
        )
        pos = torch.arange(0, maxlen).reshape(maxlen, 1)
        pos_embedding = torch.zeros((maxlen, emb_size))
        pos_embedding[:, 0::2] = torch.sin(pos * den)
        pos_embedding[:, 1::2] = torch.cos(pos * den)
        pos_embedding = pos_embedding.unsqueeze(-2)

        self.dropout = nn.Dropout(dropout)
        self.register_buffer("pos_embedding", pos_embedding)

    def forward(self, token_embedding: Tensor) -> Tensor:
        return self.dropout(  # Pos embedding behind buffer
            token_embedding + self.pos_embedding[: token_embedding.size(0), :]
        )


class ImagePositionalEncoding(nn.Module):
    """Helper Module that adds positional encoding to sequence embedding."""

    def __init__(
        self, emb_size: int, dropout: float, maxlen: int = 5000
    ) -> None:
        super().__init__()
        den = torch.exp(
            -torch.arange(0, emb_size, 2) * math.log(10000) / emb_size
        )
        pos = torch.arange(0, maxlen).reshape(maxlen, 1)
        pos_embedding = torch.zeros((maxlen, emb_size))
        pos_embedding[:, 0::2] = torch.sin(pos * den)
        pos_embedding[:, 1::2] = torch.cos(pos * den)
        pos_embedding = pos_embedding.unsqueeze(-2)

        self.dropout = nn.Dropout(dropout)
        self.register_buffer("pos_embedding", pos_embedding)

    def forward(self, token_embedding: Tensor) -> Tensor:
        return self.dropout(  # Pos embedding behind buffer
            token_embedding + self.pos_embedding[: token_embedding.size(0), :]
        )



# add stochastic ridge regression
# to do snapshot ensembling
# Possibly add simmulated annealing to determine lr


class SnapshotEnsembling:
    """Uses cosine cyclic sheduling, ~20-40 epochs per cycle."""

    pass


class FastGeometricEnsembling:
    """Uses linear piecewise cyclic sheduling, ~2-4 epochs per cycle."""

    pass


class StochasticWeightAveraging:
    """Creates two models:
    - one stores running average of model weights.
    - the other traversing weight space by cyclical learning rate.
    """



import torch.nn as nn

from aidd_codebase.utils.metacoding import DictChoiceFactory
from aidd_codebase.utils.typescripts import Tensor


class MLPChoice(DictChoiceFactory):
    pass


@MLPChoice.register_choice("mlp")
class MLP(nn.Module):
    def __init__(
        self,
        input_dim: int,
        dim_feedforward: int,
        output_dim: int,
        num_layers: int = 2,
        dropout: float = 0.0,
        add_skip_connections: bool = True,
    ) -> None:
        super().__init__()
        self.linear_net = nn.ModuleList(
            [nn.Linear(input_dim, dim_feedforward)]
        )
        self.linear_net.extend(
            [
                nn.Sequential(
                    nn.Linear(dim_feedforward, dim_feedforward),
                    nn.Dropout(dropout),
                    nn.ReLU(inplace=True),
                )
                for _ in range(1, num_layers - 1)
            ]
        )
        self.linear_net.append(nn.Linear(dim_feedforward, output_dim))

        def forward() -> Tensor:
            pass

        def residual_forward() -> Tensor:
            linear_out = self.linear_net(x)
            x = x + self.dropout(linear_out)
            x = self.norm2(x)
